@@ -23,22 +23,42 @@
  * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
  * SOFTWARE.
  */
+
+#ifdef HAVE_CONFIG_H
 #include <config.h>
+#endif
+
 #include <stdlib.h>
 #include <math.h>
 #include "pixman-private.h"
 
+static force_inline double
+coordinates_to_parameter (double x, double y, double angle)
+{
+    double t;
+
+    t = atan2 (y, x) + angle;
+
+    while (t < 0)
+	t += 2 * M_PI;
+
+    while (t >= 2 * M_PI)
+	t -= 2 * M_PI;
+
+    return 1 - t * (1 / (2 * M_PI)); /* Scale t to [0, 1] and
+				      * make rotation CCW
+				      */
+}
+
 static void
 conical_gradient_get_scanline_32 (pixman_image_t *image,
                                   int             x,
                                   int             y,
                                   int             width,
                                   uint32_t *      buffer,
-                                  const uint32_t *mask,
-                                  uint32_t        mask_bits)
+                                  const uint32_t *mask)
 {
-    source_image_t *source = (source_image_t *)image;
-    gradient_t *gradient = (gradient_t *)source;
+    gradient_t *gradient = (gradient_t *)image;
     conical_gradient_t *conical = (conical_gradient_t *)image;
     uint32_t       *end = buffer + width;
     pixman_gradient_walker_t walker;
@@ -49,11 +69,10 @@ conical_gradient_get_scanline_32 (pixman_image_t *image,
     double rx = x + 0.5;
     double ry = y + 0.5;
     double rz = 1.;
-    double a = conical->angle / (180. * 65536);
 
-    _pixman_gradient_walker_init (&walker, gradient, source->common.repeat);
+    _pixman_gradient_walker_init (&walker, gradient, image->common.repeat);
 
-    if (source->common.transform)
+    if (image->common.transform)
     {
 	pixman_vector_t v;
 
@@ -62,19 +81,19 @@ conical_gradient_get_scanline_32 (pixman_image_t *image,
 	v.vector[1] = pixman_int_to_fixed (y) + pixman_fixed_1 / 2;
 	v.vector[2] = pixman_fixed_1;
 
-	if (!pixman_transform_point_3d (source->common.transform, &v))
+	if (!pixman_transform_point_3d (image->common.transform, &v))
 	    return;
 
-	cx = source->common.transform->matrix[0][0] / 65536.;
-	cy = source->common.transform->matrix[1][0] / 65536.;
-	cz = source->common.transform->matrix[2][0] / 65536.;
-	
+	cx = image->common.transform->matrix[0][0] / 65536.;
+	cy = image->common.transform->matrix[1][0] / 65536.;
+	cz = image->common.transform->matrix[2][0] / 65536.;
+
 	rx = v.vector[0] / 65536.;
 	ry = v.vector[1] / 65536.;
 	rz = v.vector[2] / 65536.;
-	
+
 	affine =
-	    source->common.transform->matrix[2][0] == 0 &&
+	    image->common.transform->matrix[2][0] == 0 &&
 	    v.vector[2] == pixman_fixed_1;
     }
 
@@ -85,20 +104,16 @@ conical_gradient_get_scanline_32 (pixman_image_t *image,
 
 	while (buffer < end)
 	{
-	    double angle;
-
-	    if (!mask || *mask++ & mask_bits)
+	    if (!mask || *mask++)
 	    {
-		pixman_fixed_48_16_t t;
+		double t = coordinates_to_parameter (rx, ry, conical->angle);
 
-		angle = atan2 (ry, rx) + a;
-		t     = (pixman_fixed_48_16_t) (angle * (65536. / (2 * M_PI)));
-
-		*buffer = _pixman_gradient_walker_pixel (&walker, t);
+		*buffer = _pixman_gradient_walker_pixel (
+		    &walker, (pixman_fixed_48_16_t)pixman_double_to_fixed (t));
 	    }
 
 	    ++buffer;
-	    
+
 	    rx += cx;
 	    ry += cy;
 	}
@@ -108,11 +123,10 @@ conical_gradient_get_scanline_32 (pixman_image_t *image,
 	while (buffer < end)
 	{
 	    double x, y;
-	    double angle;
 
-	    if (!mask || *mask++ & mask_bits)
+	    if (!mask || *mask++)
 	    {
-		pixman_fixed_48_16_t t;
+		double t;
 
 		if (rz != 0)
 		{
@@ -126,15 +140,15 @@ conical_gradient_get_scanline_32 (pixman_image_t *image,
 
 		x -= conical->center.x / 65536.;
 		y -= conical->center.y / 65536.;
-		
-		angle = atan2 (y, x) + a;
-		t     = (pixman_fixed_48_16_t) (angle * (65536. / (2 * M_PI)));
 
-		*buffer = _pixman_gradient_walker_pixel (&walker, t);
+		t = coordinates_to_parameter (x, y, conical->angle);
+
+		*buffer = _pixman_gradient_walker_pixel (
+		    &walker, (pixman_fixed_48_16_t)pixman_double_to_fixed (t));
 	    }
 
 	    ++buffer;
-	    
+
 	    rx += cx;
 	    ry += cy;
 	    rz += cz;
@@ -169,9 +183,12 @@ pixman_image_create_conical_gradient (pixman_point_fixed_t *        center,
 	return NULL;
     }
 
+    angle = MOD (angle, pixman_int_to_fixed (360));
+
     image->type = CONICAL;
+
     conical->center = *center;
-    conical->angle = angle;
+    conical->angle = (pixman_fixed_to_double (angle) / 180.0) * M_PI;
 
     image->common.property_changed = conical_gradient_property_changed;
 
